import os
import sys
from pathlib import Path

from click.testing import CliRunner
from xsdata.cli import cli
from xsdata.models.config import GeneratorConfig
from xsdata.utils.testing import ClassFactory, FactoryTestCase

from xsdata_odoo.generator import OdooGenerator


class OdooGeneratorTests(FactoryTestCase):
    def setUp(self):
        super().setUp()
        config = GeneratorConfig()
        self.generator = OdooGenerator(config)

    def test_render(self):
        if os.environ.get("XSDATA_SCHEMA"):
            del os.environ["XSDATA_SCHEMA"]
        if os.environ.get("XSDATA_VERSION"):
            del os.environ["XSDATA_VERSION"]

        classes = [
            ClassFactory.elements(2, package="generated"),
            ClassFactory.elements(3, package="generated"),
        ]

        iterator = self.generator.render(classes)

        actual = [(out.path, out.title, out.source) for out in iterator]
        self.assertEqual(1, len(actual))
        self.assertEqual(3, len(actual[0]))
        self.assertIsInstance(actual[0][0], Path)
        self.assertTrue(actual[0][0].is_absolute())
        self.assertEqual("generated.tests", actual[0][1])
        self.assertEqual(
            str(Path("generated/tests.py")),
            str(actual[0][0].relative_to(Path.cwd())),
        )

        # this is the expected with Ruff formatting:
        expected = """# Copyright 2025 Akretion - Raphaël Valyi <raphael.valyi@akretion.com>
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html).
# Generated by https://github.com/akretion/xsdata-odoo

import textwrap
from odoo import fields, models

__NAMESPACE__ = "xsdata"


class ClassB(models.AbstractModel):
    _description = "class_B"
    _name = "spec.10.classb"
    _inherit = "spec.mixin.spec"
    _binding_type = "ClassB"



    spec10_attr_B = fields.Char(
        string="attr_B"
    )

    spec10_attr_C = fields.Char(
        string="attr_C"
    )


class ClassC(models.AbstractModel):
    _description = "class_C"
    _name = "spec.10.classc"
    _inherit = "spec.mixin.spec"
    _binding_type = "ClassC"



    spec10_attr_D = fields.Char(
        string="attr_D"
    )

    spec10_attr_E = fields.Char(
        string="attr_E"
    )

    spec10_attr_F = fields.Char(
        string="attr_F"
    )
"""
        # for convenience we remove trailing spaces:
        expected_lines = [line.rstrip() for line in expected.splitlines()]
        expected_clean = "\n".join(expected_lines)
        actual_lines = [line.rstrip() for line in actual[0][2].splitlines()]
        actual_clean = "\n".join(actual_lines)
        self.assertEqual(expected_clean, actual_clean)

    def test_complete_po(self):
        os.environ["XSDATA_SCHEMA"] = "poxsd"
        if os.environ.get("XSDATA_VERSION"):
            del os.environ["XSDATA_VERSION"]

        runner = CliRunner()
        schema = Path(__file__).parent.joinpath("fixtures/po/po.xsd")
        os.chdir(Path(__file__).parent.parent)

        result = runner.invoke(
            cli,
            [
                str(schema),
                "--package",
                "generated.po.models",
                "--structure-style=single-package",
                "--output",
                "odoo",
            ],
            catch_exceptions=True,
        )

        self.assertIsNone(result.exception)

        if "win" not in sys.platform.lower():
            expected = "to be read 1"
            generated = "to be read 2"
            with open("tests/fixtures/po/models.py") as f:
                expected = f.read()
            with open("generated/po/models.py") as f:
                generated = f.read()
            self.assertEqual(expected, generated)

    def test_complete_po_with_nesting(self):
        if os.environ.get("XSDATA_SCHEMA"):
            del os.environ["XSDATA_SCHEMA"]
        if os.environ.get("XSDATA_VERSION"):
            del os.environ["XSDATA_VERSION"]

        runner = CliRunner()
        schema = Path(__file__).parent.joinpath("fixtures/po_with_nesting/po.xsd")
        os.chdir(Path(__file__).parent.parent)

        result = runner.invoke(
            cli,
            [
                str(schema),
                "--package",
                "generated.po_with_nesting.models",
                "--structure-style=single-package",
                "--output",
                "odoo",
            ],
            catch_exceptions=True,
        )

        self.assertIsNone(result.exception)

        if "win" not in sys.platform.lower():
            expected = "to be read 1"
            generated = "to be read 2"
            with open("tests/fixtures/po_with_nesting/models.py") as f:
                expected = f.read()
            with open("generated/po_with_nesting/models.py") as f:
                generated = f.read()
            self.assertEqual(expected, generated)

    def test_complete_nfe(self):
        os.environ["XSDATA_SCHEMA"] = "nfe"
        os.environ["XSDATA_VERSION"] = "40"
        os.environ["XSDATA_SKIP"] = r"^ICMS.ICMS\d+|^ICMS.ICMSSN\d+"
        os.environ["XSDATA_LANG"] = "portuguese"

        runner = CliRunner()
        schema = Path(__file__).parent.joinpath("fixtures/nfe/leiauteNFe_v4.00.xsd")
        os.chdir(Path(__file__).parent.parent)

        result = runner.invoke(
            cli,
            [
                str(schema),
                "--package",
                "generated.nfe.v4_0.models",
                "--structure-style=single-package",
                "--output",
                "odoo",
            ],
            catch_exceptions=True,
        )

        self.assertIsNone(result.exception)

        if sys.version_info[:2] > (3, 9) and "win" not in sys.platform.lower():
            expected = "to be read 1"
            generated = "to be read 2"
            with open("tests/fixtures/nfe/models.py") as f:
                expected = f.read()
            with open("generated/nfe/v4_0/models.py") as f:
                generated = f.read()
            self.assertEqual(expected, generated)



import os
from typing import List

from xsdata.codegen.handlers.merge_attributes import MergeAttributes
from xsdata.codegen.handlers.update_attributes_effective_choice import (
    UpdateAttributesEffectiveChoice,
)
from xsdata.codegen.models import Attr, Class
from xsdata.codegen.utils import ClassUtils
from xsdata.codegen.writer import CodeWriter
from xsdata.utils import collections

from xsdata_odoo.generator import OdooGenerator


@classmethod  # type: ignore[misc]  # Suppresses "classmethod used with a non-method"
def process(cls, target: Class):
    """
    Detect same type attributes in order to merge them together with their
    restrictions.

    Two attributes are considered equal if they have the same name,
    tag and namespace.
    """
    result: List[Attr] = []
    for attr in target.attrs:
        pos = collections.find(result, attr)
        existing = result[pos] if pos > -1 else None

        if not existing:
            result.append(attr)
        elif not (attr.is_attribute or attr.is_enumeration):
            existing.help = existing.help or attr.help

            e_res = existing.restrictions
            a_res = attr.restrictions

            min_occurs = e_res.min_occurs or 0
            max_occurs = e_res.max_occurs or 1
            attr_min_occurs = a_res.min_occurs or 0
            attr_max_occurs = a_res.max_occurs or 1

            e_res.min_occurs = min(min_occurs, attr_min_occurs)
            if os.environ.get("XSDATA_SCHEMA") in ("nfe",):
                e_res.max_occurs = min(max_occurs, attr_max_occurs)  # this is the patch
            else:
                e_res.max_occurs = max_occurs + attr_max_occurs
            e_res.sequential = a_res.sequential or e_res.sequential
            existing.fixed = False
            existing.types.extend(attr.types)

    target.attrs = result
    ClassUtils.cleanup_class(target)


@classmethod  # type: ignore[misc]  # Suppresses "classmethod used with a non-method"
def merge_attrs(cls, target: Class, groups: List[List[int]]) -> List[Attr]:
    attrs = []

    for index, attr in enumerate(target.attrs):
        group = collections.find_connected_component(groups, index)

        if group == -1:
            attrs.append(attr)
            continue

        pos = collections.find(attrs, attr)
        if pos == -1:
            attr.restrictions.choice = (group * -1) - 1
            attrs.append(attr)
        else:
            existing = attrs[pos]
            assert existing.restrictions.min_occurs is not None
            assert existing.restrictions.max_occurs is not None

            existing.restrictions.min_occurs += attr.restrictions.min_occurs or 0
            if os.environ.get("XSDATA_SCHEMA") in ("nfe",) and attr.name == "IPI":
                existing.restrictions.max_occurs = min(
                    attr.restrictions.max_occurs or 0,
                    existing.restrictions.max_occurs,
                )
            else:
                existing.restrictions.max_occurs += attr.restrictions.max_occurs or 0

    return attrs


# workaround for the Brazilian NFe https://github.com/akretion/nfelib/issues/40
# another option would be to use the --compound-fields option but that would
# force us to rework a bit the spec_driven_model module logic a bit.
# see https://github.com/akretion/nfelib/issues/40
if hasattr(UpdateAttributesEffectiveChoice, "merge_attrs"):
    # xsdata > 22.12
    merge_attrs._original_method = UpdateAttributesEffectiveChoice.merge_attrs  # type: ignore[attr-defined]
    UpdateAttributesEffectiveChoice.merge_attrs = merge_attrs
else:
    process._original_method = MergeAttributes.process  # type: ignore[attr-defined]
    MergeAttributes.process = process

CodeWriter.register_generator("odoo", OdooGenerator)



import os
import re
from collections import OrderedDict
from typing import Any, Dict, List, Optional

from jinja2 import Environment
from xsdata.codegen.models import Attr, Class
from xsdata.formats.dataclass.filters import Filters
from xsdata.logger import logger
from xsdata.models.config import GeneratorConfig, ObjectType
from xsdata.utils import namespaces

from .text_utils import extract_string_and_help, wrap_text

INTEGER_TYPES = ("integer", "positiveInteger")
FLOAT_TYPES = ("float", "decimal")
# Odoo has no Decimal field and all in all it's better to guess float by default
# and override if required.
# DECIMAL_TYPES = ("decimal",)
MONETARY_DIGITS = 2
CHAR_TYPES = (
    "string",
    "NMTOKEN",
    "ID",
    "IDREF",
    "IDREFS",
    "anyURI",
    "base64Binary",
    "normalizedString",
    "language",
)
DATE_TYPES = ("date", "TData")
DATETIME_TYPES = ("dateTime", "TDateTimeUTC")
BOOLEAN_TYPES = "boolean"

# generally it's not interresting to generate mixins for signature
SIGNATURE_CLASS_SKIP = [
    "^Signature$",
    "^SignatureType$",
    "^SignatureValueType$",
    "^SignedInfoType$",
    "^ReferenceType$",
    "^DigestMethodType$",
    "^TransformsType$",
    "^TransformType$",
    "^KeyInfoType$",
    "^X509DataType$",
    "^CanonicalizationMethodType$",
    "^SignatureMethodType$",
    "^RSAKeyValue$",
    "^TRSAKeyValueType$",
]


class OdooFilters(Filters):
    __slots__ = (
        "files_to_etree",
        "all_simple_types",
        "all_complex_types",
        "registry_names",
        "implicit_many2ones",
        "schema",
        "version",
        "python_inherit_model",
        "inherit_model",
        "xsd_extra_info",
        "relative_imports",
    )

    def __init__(
        self,
        config: GeneratorConfig,
        all_simple_types: List[Class],
        all_complex_types: List[Class],
        registry_names: Dict,
        implicit_many2ones: Dict,
        schema: str = "spec",
        version: str = "10",
        python_inherit_model: str = "models.AbstractModel",
        inherit_model=None,
    ):
        super().__init__(config)
        self.all_simple_types = all_simple_types
        self.all_complex_types = all_complex_types
        self.registry_names = registry_names
        self.implicit_many2ones = implicit_many2ones
        self.files_to_etree: Dict[str, Any] = {}
        self.schema = schema
        self.version = version
        self.python_inherit_model = python_inherit_model
        if inherit_model is None:
            inherit_model = f"spec.mixin.{schema}"
        self.inherit_model = inherit_model
        self.xsd_extra_info: Dict[str, Any] = {}
        self.relative_imports = True

    def register(self, env: Environment):
        super().register(env)
        env.filters.update(
            {
                "enum_skip": self.enum_skip,
                "pattern_skip": self.pattern_skip,
                "registry_name": self.registry_name,
                "odoo_python_inherit_model": self.odoo_python_inherit_model,
                "odoo_inherit_model": self.odoo_inherit_model,
                "clean_docstring": self.clean_docstring,
                "binding_type": self.binding_type,
                "odoo_class_name": self.odoo_class_name,
                "class_properties": self.class_properties,
                "odoo_class_description": self.odoo_class_description,
                "odoo_field_definition": self.odoo_field_definition,
                "odoo_implicit_many2ones": self.odoo_implicit_many2ones,
                "import_class": self.import_class,
                "enum_docstring": self.enum_docstring,
            }
        )

    def enum_skip(self, obj: Class, name: str) -> bool:
        """
        Avoids Postgres errors with fields.Selection
        by disabling case insensitive duplicates.
        For instance:
        TENDEREMI_XPAIS = [
            ("Brasil", "Brasil"),
        # ("BRASIL", "BRASIL"),
        ]
        """
        lower_vals = set()
        allowed_vals = set()
        for attr in obj.attrs:
            if attr.default.lower() not in lower_vals:
                lower_vals.add(attr.default.lower())
                allowed_vals.add(attr.default)
        if name in allowed_vals:
            return False
        else:
            return True

    def pattern_skip(self, name: str, parents: Optional[List[Class]] = None) -> bool:
        """Should class or field be skipped?"""
        if parents is None:
            parents = []
        class_skip = SIGNATURE_CLASS_SKIP.copy()
        if os.environ.get("XSDATA_SKIP"):
            class_skip += os.environ["XSDATA_SKIP"].split("|")
        for pattern in class_skip:
            # do we have a simple match? (no scoping can be risky)
            if re.search(pattern, name):
                return True
            part_count = pattern.count(".") + 1
            if part_count > 1:
                # eventually we search for the class with its parents scope
                parent_pattern = ".".join(
                    [namespaces.local_name(c.qname) for c in parents[-part_count:]]
                )
                if re.search(pattern, parent_pattern):
                    return True
                # we now search for the field_name with its parents scope
                field_parent_pattern = ".".join(
                    [namespaces.local_name(c.qname) for c in parents[-part_count + 1 :]]
                    + [name]
                )
                if re.search(pattern, field_parent_pattern):
                    return True

        return False

    def odoo_class_description(self, obj: Class) -> str:
        if obj.help:
            return """textwrap.dedent(f"    {__doc__}")"""
        else:
            # TODO some inner classes have no obj.help
            # but we can read the XML annotations. Ex: "TretEnviNfe.InfRec", "Tnfe.InfNfe"...
            return f'"{obj.name}"'

    def enum_docstring(self, obj: Class) -> str:
        """Also format enum items. Works well for Brazilian fiscal xsd, may need adaptations for
        your xsd."""
        separators = (" - ", "-", " – ", "–")
        if "_" in obj.name:
            type_qname = obj.qname.split("_")[0]
            field_name = obj.name.split("_")[1]
            for klass in self.all_complex_types:
                if klass.qname == type_qname:
                    for idx, item in enumerate(obj.attrs):
                        for field in klass.attrs:
                            if field.name != field_name or not field.help:
                                continue
                            item_help = False
                            for separator in separators:
                                split = field.help.split(f"{item.default}{separator}")
                                if len(split) > 1:
                                    # TODO sometimes the line may continue
                                    # until the next value or may end at next value...
                                    item_help = split[1].splitlines()[0].split(";")[0]
                                    break

                            if item_help:
                                item.help = item_help
                                if idx == 0 and len(split) > 1:
                                    obj.help, _help_trash = extract_string_and_help(
                                        obj.name,
                                        field.name,
                                        split[0],
                                        set(),
                                        1024,
                                    )
                            else:
                                item.help = item.default
                            if (
                                idx == 0
                                and not obj.help
                                and not field.help.startswith(item.default)
                            ):
                                obj.help = (
                                    field.help.strip()
                                )  # no split but better than no docstring

        for item in obj.attrs:  # (it also apply to simple_types)
            if not item.help:
                item.help = f'"{item.default}"'
                continue
            item.help = item.help.replace("\n", "").replace('"', "'").strip()
            if len(item.help) > 78:
                lines = [
                    f'"{line.strip()} "'.replace('- "', '-"')
                    for line in wrap_text(item.help, 8, 78)
                    .replace('"""', "")
                    .splitlines()
                ]
                lines[-1] = lines[-1].replace(' "', '"')
                lines_str = "\n     ".join(lines)
                item.help = f"\n    ({lines_str})"
            else:
                item.help = f'"{item.help}"'

        if obj.help:
            return "# " + "\n# ".join(
                wrap_text(obj.help.strip(), 0, 78).replace('"', "").splitlines()
            )
        else:
            return ""

    def odoo_class_name(self, obj: Class, parents: List[Class] = []):
        """
        Same as class_name with the --unnest-classes option but without the
        option side effects.
        """
        full_name = ".".join([self.class_name(c.name) for c in parents])
        return self.registry_names[full_name].replace(".", "")

    def registry_name(
        self,
        name: str = "",
        parents: List[Class] = [],
        type_names: List[str] = [],
    ) -> str:
        if parents:
            full_name = ".".join([self.class_name(c.name) for c in parents])
        else:
            if type_names:
                full_name = ".".join(type_names)
            else:
                full_name = self.class_name(name)
        # NOTE we cannot use the class ref as a key because only type names
        # are provided by xsdata for fields
        unique_name = self.registry_names[full_name].replace(".", "_")
        return f"{self.schema}.{self.version}.{unique_name.lower()}"

    def odoo_inherit_model(self, obj: Class) -> str:
        return self.inherit_model

    def odoo_python_inherit_model(self, obj: Class) -> str:
        return self.python_inherit_model

    def registry_comodel(self, type_name: str):
        # NOTE: we take only the last part of inner Types with .split(".")[-1]
        # but if that were to create Type duplicates we could change that.
        clean_type_names = type_name.replace('"', "").split(".")
        return self.registry_name(clean_type_names[-1], type_names=clean_type_names)

    def clean_docstring(self, string: Optional[str], escape: bool = True) -> str:
        """Prepare string for docstring generation."""
        if not string:
            return ""  # TODO read from parent field if any
        return "\n    {}".format(wrap_text(string.strip(), 4, 79))

    def class_properties(
        self,
        obj: Class,
        parents: List[Class],
    ) -> str:
        """Return the name of the xsdata class for a given Odoo model."""
        if os.environ.get("XSDATA_GENDS"):
            return (
                f'_binding_type = "{self.binding_type(obj, parents)}"\n'
                f'    _generateds_type = "{self.generateds_type(obj, parents)}"'
            )
        else:
            return f'_binding_type = "{self.binding_type(obj, parents)}"\n'

    def binding_type(
        self,
        obj: Class,
        parents: List[Class],
    ) -> str:
        """Return the name of the xsdata class for a given Odoo model."""
        return ".".join([self.class_name(p.name) for p in parents])

    def generateds_type(
        self,
        obj: Class,
        parents: List[Class],
    ) -> str:
        """
        DEPRECATED!
        Return the name of the GenerateDS class for a given Odoo model.
        This is for backward compatibility: it allows to use Odoo mixins
        generated with xsdata along with legacy GenerateDS Python bindings.
        """
        if len(parents) > 1:
            return obj.qname.split("}")[1] + "Type"
        else:
            return obj.qname.split("}")[1]

    def odoo_implicit_many2ones(self, obj: Class, parents: List[Class]) -> str:
        """The m2o fields for the o2m keys."""
        fields = []
        implicit_many2ones = self.implicit_many2ones.get(
            self.binding_type(obj, parents).lower(),
            [],
            # NOTE: strangely lower is required (Brazilian CTe)
        )
        for implicit_many2one_data in implicit_many2ones:
            kwargs = OrderedDict()
            kwargs["comodel_name"] = implicit_many2one_data[0]
            kwargs["xsd_implicit"] = True
            # kwargs["required"] = True  # FIXME seems it creates ORM issues
            kwargs["ondelete"] = "cascade"
            target_field = implicit_many2one_data[1]
            fields.append(
                f"{target_field} = fields.Many2one({self.format_arguments(kwargs, 4)})"
            )
        return ("\n").join(fields)

    def field_name(self, name: str, class_name: str) -> str:
        """Like xsdata but you can enforce the prefix or safe_name
        conversion."""
        prefix = self.field_safe_prefix
        name = self.apply_substitutions(name, ObjectType.FIELD)

        if self.field_safe_prefix in ("", "value"):  # do like standard xsdata
            name = self.safe_name(name, prefix, self.field_case, class_name=class_name)
        if self.field_safe_prefix == "NO_PREFIX_NO_SAFE_NAME":  # keep my field name!
            pass
        elif self.field_safe_prefix.endswith("SAFE_NAME"):  # prefix + python field name
            prefix = prefix.split("SAFE_NAME")[0]
            name = f"{prefix}{name}"
            name = self.safe_name(name, prefix, self.field_case, class_name=class_name)
        else:  # prefix + keep my field name
            name = f"{prefix}{name}"

        return self.apply_substitutions(name, ObjectType.FIELD)

    def odoo_field_definition(
        self,
        attr: Attr,
        parents: List[Class],
    ) -> str:
        """Return the Odoo field definition."""

        # 1st some checks inspired from xsdata Filters:
        obj = parents[-1]
        type_names = self._field_type_names(obj, attr)
        # collections.unique_sequence(
        #    self._field_type_name(x, [p.name for p in parents]) for x in attr.types
        # )
        if len(attr.types) > 1:
            logger.warning(
                f"len(type_names) > 1 (Union) not implemented yet! class: {obj.name} attr: {attr}"
            )

        if attr.is_tokens:
            logger.warning(
                f"attr.is_tokens not implemented yet! class: {obj.name} attr: {attr}"
            )

        if attr.is_dict:
            logger.warning(
                f"attr.is_dict not implemented yet! class: {obj.name} attr: {attr}"
            )

        # default_value = self.field_default_value(attr, {})

        kwargs = self._extract_field_attributes(parents, attr)

        if attr.types[0].datatype:  # simple field
            return self._simple_field_definition(obj, attr, type_names, kwargs)

        else:  # relational field
            field = self._try_one2many_field_definition(obj, attr, type_names, kwargs)
            if field is None:
                field = self._try_selection_field_definition(
                    obj, attr, type_names, kwargs
                )
                if field is None:
                    field = self._try_many2one_field_definition(
                        obj, attr, type_names, kwargs
                    )

            if field is not None:
                return field

            message = f"Missing class {attr.types[0]}! class: {obj.name} attr: {attr}"
            logger.warning(message)
            return message

    def _extract_field_attributes(
        self, parents: List[Class], attr: Attr
    ) -> OrderedDict[str, Any]:
        obj = parents[-1]
        kwargs: OrderedDict[str, Any] = OrderedDict()
        if not hasattr(obj, "unique_labels"):
            obj.unique_labels = set()  # will avoid repeating field labels
        string, help_attr = extract_string_and_help(
            obj.name, attr.name, attr.help, obj.unique_labels
        )
        kwargs["string"] = string

        metadata = self.field_metadata(obj, attr, None)
        if metadata.get("required"):
            # we choose not to put required=True (required in database) to avoid
            # messing with existing Odoo modules.
            kwargs["xsd_required"] = True

        kwargs.update(self.xsd_extra_info.get(f"{obj.name}#{attr.name}", {}))
        if help_attr and not kwargs.get("help"):
            kwargs["help"] = help_attr  # (help as the last attribute)

        return kwargs

    def _extract_number_attrs(
        self, obj: Class, attr: Attr, kwargs: OrderedDict[str, Any]
    ):
        """
        Monetary vs Float field detection.

        Detection tends to be brittle but in general it doesn't impact
        XML serialization/desrialization as both types are floats and we
        usually take the xsd_type into account.

        You can somewhat customize the default behavior with ENV VARs:
        XSDATA_MONETARY_TYPE: xsd type to force fields.Monetary
        XSDATA_NUM_TYPE: xsd type for fields.Float or fields.Monetary eventually
        you can use a prefix followed by brackets to indicate where to take the
        floor part and the decimal part like Prefix[in_start:int_stop.dec_start:dec_stop]
        """
        python_type = attr.types[0].datatype.code
        if python_type in FLOAT_TYPES or python_type in CHAR_TYPES:
            xsd_type = kwargs.get("xsd_type", "")

            monetary_type = os.environ.get("XSDATA_MONETARY_TYPE")

            num_type_complete = os.environ.get("XSDATA_NUM_TYPE", "TDec_[5:7.7:9]")
            if "[" in num_type_complete:
                int_part = num_type_complete.split("[")[1].split(".")[0]
                int_start = int(int_part.split(":")[0])
                int_stop = int(int_part.split(":")[1])
                dec_part = num_type_complete.split(".")[1].split("]")[0]
                dec_start = int(dec_part.split(":")[0])
                dec_stop = int(dec_part.split(":")[1])
                num_type = num_type_complete.split("[")[0]
                conditional_monetary = True
            else:
                num_type = num_type_complete
                conditional_monetary = False

            if monetary_type and xsd_type.startswith(monetary_type):
                kwargs["currency_field"] = "brl_currency_id"  # TODO use spec_curreny_id
            elif xsd_type.startswith(num_type):
                if conditional_monetary:
                    if int(
                        xsd_type.replace("03v", "03")[dec_start:dec_stop]
                    ) != MONETARY_DIGITS or (
                        # for Brazilian edocs, pSomething means percentualSomething ->Float
                        attr.name[0] == "p" and attr.name[1].isupper()
                    ):
                        kwargs["digits"] = (
                            int(xsd_type.replace("03v", "03")[int_start:int_stop]),
                            int(xsd_type.replace("03v", "03")[dec_start:dec_stop]),
                        )
                    else:
                        kwargs["currency_field"] = (
                            "brl_currency_id"  # TODO use spec_curreny_id
                        )
                else:
                    kwargs["digits"] = (16, 4)

    def _simple_field_definition(
        self, obj: Class, attr: Attr, type_names: str, kwargs: OrderedDict
    ):
        self._extract_number_attrs(obj, attr, kwargs)
        if kwargs.get("help"):
            kwargs.move_to_end("help", last=True)
        python_type = attr.types[0].datatype.code
        if python_type in DATE_TYPES or kwargs.get("xsd_type") in DATE_TYPES:
            return f"fields.Date({self.format_arguments(kwargs, 4)})"
        elif python_type in DATETIME_TYPES or kwargs.get("xsd_type") in DATETIME_TYPES:
            return f"fields.Datetime({self.format_arguments(kwargs, 4)})"
        elif python_type in INTEGER_TYPES:
            return f"fields.Integer({self.format_arguments(kwargs, 4)})"
        if kwargs.get("currency_field"):
            return f"fields.Monetary({self.format_arguments(kwargs, 4)})"
        elif python_type in FLOAT_TYPES or kwargs.get("digits"):
            return f"fields.Float({self.format_arguments(kwargs, 4)})"
        elif python_type in CHAR_TYPES:
            return f"fields.Char({self.format_arguments(kwargs, 4)})"
        elif python_type in BOOLEAN_TYPES:
            return f"fields.Boolean({self.format_arguments(kwargs, 4)})"
        else:
            logger.warning(
                f"{python_type} {attr.types[0].datatype} not implemented yet! class: {obj.name} attr: {attr}"
            )
            return ""

    def _try_one2many_field_definition(
        self, obj: Class, attr: Attr, type_names: str, kwargs: OrderedDict
    ):
        if attr.is_list:
            if self.pattern_skip(attr.types[0].name):
                return ""
            comodel_key = self.field_name(f"{attr.name}_{obj.name}_id", obj.name)
            if type_names.split(".")[-1].lower().replace('"', "") not in [
                t.name.lower() for t in self.all_complex_types
            ]:
                logger.warning(
                    f"no complex type found for {type_names}; skipping attr {attr.name} in class {obj.name}!"
                )
                # example cte40_cInfManu in Brazilian CTe. Seems like a o2m to a simple type/Enum. Not implemented yet.
                return ""
            return f"""fields.One2many("{self.registry_comodel(type_names)}", "{comodel_key}",{self.format_arguments(kwargs, 4)})"""

    def _try_selection_field_definition(
        self, obj: Class, attr: Attr, type_names: str, kwargs: OrderedDict
    ):
        for klass in self.all_simple_types:
            if attr.types[0].qname == klass.qname:
                if self.pattern_skip(klass.name.upper()):
                    return ""

                return f"fields.Selection({klass.name.upper()},{self.format_arguments(kwargs, 4)})"

    def _try_many2one_field_definition(
        self, obj: Class, attr: Attr, type_names: str, kwargs: OrderedDict
    ):
        for klass in self.all_complex_types:
            if attr.types[0].qname == klass.qname:
                if self.pattern_skip(attr.types[0].name):
                    return ""
                # Many2one
                kwargs["comodel_name"] = self.registry_comodel(type_names)
                kwargs.move_to_end("comodel_name", last=False)
                return f"fields.Many2one({self.format_arguments(kwargs, 4)})"

    def import_class(self, name: str, alias: Optional[str]) -> str:
        """Convert import class name with alias support."""
        if alias:
            return f"{self.class_name(name)} as {self.class_name(alias)}"
        if name in [klass.name for klass in self.all_simple_types]:
            return name.upper()  # const are upcase in Odoo
        else:
            return self.class_name(name)



import locale
import os
import textwrap
from typing import Tuple

import babel
from nltk import download
from nltk.corpus import stopwords

# where to stop when trying to extract the beginning of a text
STRONG_PONCT_TOKENS = (". ", ", ", " (", " - ", ".", ",", ": ", "|")
STRING_MIN_LEN = 36  # agressive cuts
STRING_MAX_LEN = 40  # progressive cuts

USELESS_STARTS = ("Informar o ", "Informar a ", "Preencher com ")


def extract_string_and_help(
    obj_name: str,
    attr_name: str,
    doc: str,
    unique_labels: set,
    max_len: int = STRING_MAX_LEN,
) -> Tuple[str, str]:
    """
    Try to extract a proper field string from any (xsd) longest field documentation.
    Eventually attr_name is technical and a better string/label can be
    extracted from the beginning of the help text.
    Giving good results 95% of the time is fine because one can override
    in Odoo fields in the 5% of cases where the extracted string isn't nice.
    Labels duplicates are avoided using the unique_labels set param and by
    appending the field name in the description eventually if required.
    """

    string = attr_name
    if doc:
        if "Tomador do Serviço" in doc and "Preencher com" in doc:
            # FIXME hack for Brazilian CTe:
            doc = doc.strip().replace('"', "'").replace("\\", "/").replace("\t", " ")
        else:
            doc = doc.strip().replace('"', "'")
        string, doc = _aggressive_cut(doc)
        string = _progressive_cut(string, max_len)

        if "(" in string and ")" not in string:
            string = string.split("(")[0].strip()
        if len(string) > max_len:
            string = attr_name

        if string == doc or doc[:-1] == string:  # doc might end with '.'
            doc = ""

    if string in unique_labels:
        string = f"{string} ({attr_name})"
        if len(string) > max_len:
            string = attr_name

    unique_labels.add(string)
    return string, doc


def _aggressive_cut(doc):
    """
    Use ponctuation to cut the string.
    Remove non relevant starting strings from a field string.
    non relevant starting strings can be forced through the USELESS_STARTS env var.
    """

    def remove_after(string, token):
        return token.join(string.split(token)[:-1])

    if os.environ.get("USLESS_STARTS"):
        useless_starts = tuple(os.environ["USELESS_STARTS"].split("|"))
    else:
        useless_starts = USELESS_STARTS  # a good default for us in Brazil
    for start in useless_starts:
        if doc.lower().startswith(start.lower()):
            doc = doc[len(start) :]

    string = " ".join(doc.splitlines()[0].split())  # avoid double spaces

    for token in STRONG_PONCT_TOKENS:  # cut aggressively
        while token in string:
            if len(string) > STRING_MIN_LEN:
                string = remove_after(string, token)
            else:
                break

    return string, doc


def _progressive_cut(string, max_len):
    """
    Try to cut the field string progressively using language specific stopwords.
    language is detected with locale.getdefaultlocale() but can be forced
    through the LANGUAGE ENV var. This will work well for some languages like
    Latin languages. For some other languages (German?), may be customization
    would be required.
    """
    if os.environ.get("XSDATA_LANG"):
        lang = os.environ["XSDATA_LANG"]
    else:
        os_locale = locale.getdefaultlocale()
        lang = (
            babel.Locale.parse(os_locale[0])
            .get_display_name("en")
            .split("(")[0]
            .lower()
            .strip()
        )
    try:
        lang_stopwords = stopwords.words(lang)
    except LookupError:
        download("stopwords")
        lang_stopwords = stopwords.words(lang)

    while len(string) > max_len:
        max_index = 0
        max_token = None
        for stopword in lang_stopwords:  # cut progressively
            token = f" {stopword} "
            if token in string and string.rindex(token) > max_index:
                max_index = string.rindex(token)
                max_token = token
        if max_token:
            string = max_token.join(string.split(max_token)[:-1])
        else:
            break

    should_try_cut = True
    while should_try_cut:
        for token in STRONG_PONCT_TOKENS + tuple(
            [f" {w} " for w in stopwords.words(lang)]
        ):
            if string.endswith(token.rstrip()):
                string = string[: string.rindex(token.rstrip())]
                should_try_cut = True
                break
            else:
                should_try_cut = False

    return string


def wrap_text(
    text,
    indent,
    width=88,
    initial_indent=4,
):
    text = text.strip()
    if len(text) + initial_indent + 8 > width or "\n" in text or '"' in text:
        quote = '"""'
    else:
        quote = '"'
    if text[0] == '"':
        text = f"{quote} {text}"
    else:
        text = f"{quote}{text}"
    if text[-1] == '"':
        text = f"{text} {quote}"
    else:
        text = f"{text}{quote}"

    wrapped_lines = []
    first = True
    for line in text.splitlines():
        if first:
            w = width - initial_indent
            first = False
        else:
            w = width
        lines = textwrap.fill(
            " ".join(line.strip().split()),
            width=w,
            subsequent_indent=" " * indent,
            replace_whitespace=False,
        ).splitlines()
        wrapped_lines += [i.strip() for i in lines]
    text = ("\n" + " " * indent).join(wrapped_lines)

    return text



import os
import re
import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

from jinja2 import Environment, FileSystemLoader
from lxml import etree
from xsdata.codegen.models import Class
from xsdata.codegen.resolver import DependenciesResolver
from xsdata.codegen.exceptions import CodegenError
from xsdata.formats.dataclass.generator import DataclassGenerator
from xsdata.formats.mixins import GeneratorResult
from xsdata.models.config import GeneratorConfig
from xsdata.utils import collections

from .codegen.resolver import OdooDependenciesResolver
from .filters import SIGNATURE_CLASS_SKIP, OdooFilters

# only put this header in files with complex types (not in tipos_basico_v4_00.py for instance)
# import textwrap
# from odoo import fields, models

# WISHLIST pluggable filters (test with UBL and cbc: => simpleType + UBL simple types mapping)
# (see UBL branch for UBL)


class OdooGenerator(DataclassGenerator):
    """Odoo generator."""

    def __init__(self, config: GeneratorConfig):
        schema = os.environ.get("XSDATA_SCHEMA", "spec")
        version = os.environ.get("XSDATA_VERSION", "10")
        # if field prefix is not set via the config (default is "value")
        # then set it with SCHEMA and VERSION env vars
        if config.conventions.field_name.safe_prefix == "value":
            config.conventions.field_name.safe_prefix = f"{schema}{version}_"

        super().__init__(config)
        self.all_simple_types: List[Class] = []
        self.all_complex_types: List[Class] = []
        self.registry_names: Dict = {}
        self.implicit_many2ones: Dict = defaultdict(list)
        tpl_dir = Path(__file__).parent.joinpath("templates")
        self.env = Environment(loader=FileSystemLoader(str(tpl_dir)), autoescape=False)
        self.filters = OdooFilters(
            config,
            self.all_simple_types,
            self.all_complex_types,
            self.registry_names,
            self.implicit_many2ones,
            schema,
            version,
        )
        self.filters.register(self.env)

    def _find_duplicated_names(self, class_paths):
        duplicates = set()
        names = set()

        for ref, path in class_paths.items():
            path_parts = path.split("|")
            name = path_parts[-1]
            if name in names:
                duplicates.add(name)
            names.add(name)

        return duplicates

    def simplify_name_sets(self, names_dict, class_names):
        """
        Find minimal unique name for class. We use dict keys as an ordered set.
        """
        initial_length = len(names_dict.keys())
        max_positions = max(map(lambda i: len(i.split("|")), names_dict.keys()))
        i = max_positions - 2
        while i >= 0:
            # try to remove each path part while ensuring no path collision (shorter len)
            test = {
                k: ""
                for k in map(
                    lambda path: len(path.split("|")) > i + 1
                    and "|".join(path.split("|")[:i] + path.split("|")[i + 1 :])
                    not in class_names
                    and "|".join(path.split("|")[:i] + path.split("|")[i + 1 :])
                    or path,
                    names_dict.keys(),
                )
            }
            i -= 1
            if len(test.keys()) == initial_length:
                names_dict = test
        return names_dict

    def _find_minimal_unique_name(self, class_paths, path_parts):
        """
        Find minimal unique name for class
        Minimal unique name is found by prepending the first uncommon parent to class name.
        """

        name = path_parts[-1]
        orig_path = "|".join(path_parts)
        duplicates = list()

        for ref, path in class_paths.items():
            if path.endswith("|" + name) and path != orig_path:
                duplicates.append(path)  # .split("|"))

        res = self.simplify_name_sets(
            {k: "" for k in [orig_path] + duplicates}, class_paths.values()
        )
        return list(res.keys())[0].replace("|", ".")

    def _find_minimal_unique_names(self, class_paths, duplicates):
        """
        Find minimal unique names for classes
        """

        class_names = dict()
        for ref, path in class_paths.items():
            path_parts = path.split("|")
            name = path_parts[-1]
            if name in duplicates:
                name = self._find_minimal_unique_name(class_paths, path_parts)

            class_names[ref] = name
        return class_names

    def _generate_unique_class_names(self, class_paths):
        """
        Generate unique class name
        Class names must be unique in the module, so we find the minimum
        unique name for each class using a depth-first search and
        prepending the parent class name.
        """
        duplicates = self._find_duplicated_names(class_paths)
        class_names = self._find_minimal_unique_names(class_paths, duplicates)
        return {
            ".".join([self.filters.class_name(i) for i in v.split("|")]): ".".join(
                [self.filters.class_name(name) for name in class_names[k].split(".")]
            )
            for k, v in class_paths.items()
        }

    def render(self, classes: List[Class]) -> Iterator[GeneratorResult]:
        """Return a iterator of the generated results."""
        registry = {obj.qname: obj.target_module for obj in classes}
        resolver = OdooDependenciesResolver(registry=registry)

        # Generate packages
        for path, cluster in self.group_by_package(classes).items():
            yield from self.ensure_packages(path.parent)

            class_paths = dict()

            def dfs(visited, graph, node, path=""):
                if (node, path) not in visited:
                    visited.append((node, path))
                    if path:
                        path = path + f"|{node.name}"
                    else:
                        path = node.name
                    class_paths[node.ref] = path
                    for neighbour in node.inner:  # in graph[node]:
                        # FIXME is this parent collecting buggy??
                        dfs(visited, graph, neighbour, path)

            all_file_classes: List[Any] = []
            for c in cluster:
                dfs(all_file_classes, cluster, c)  # , c.name)

            unique_class_names = self._generate_unique_class_names(class_paths)
            for _ref, name in class_paths.items():
                full_name = ".".join(
                    [self.filters.class_name(i) for i in name.split("|")]
                )
                self.registry_names[full_name] = unique_class_names[full_name]

            for klass, path in all_file_classes:
                if not self.filters.files_to_etree.get(
                    klass.location
                ) and os.path.isfile(klass.location):
                    xsd_tree = etree.parse(klass.location)
                    self.filters.files_to_etree[klass.location] = xsd_tree

                if klass.is_enumeration and klass not in self.all_simple_types:
                    self.all_simple_types.append(
                        klass
                    )  # TODO add module name/path for import?
                for field in klass.attrs:
                    if not field.types[0].datatype and field.is_list:
                        if path:
                            parent_names = [
                                self.filters.class_name(i) for i in path.split("|")
                            ] + [self.filters.class_name(klass.name)]
                        else:
                            parent_names = [self.filters.class_name(klass.name)]

                        type_names = collections.unique_sequence(
                            self.filters._field_type_name(klass, x) for x in field.types
                        )
                        if self.filters.pattern_skip(type_names[0]):
                            continue
                        target_field = self.filters.field_name(
                            f"{field.name}_{klass.name}_id",
                            klass.name,
                        )
                        comodel_type = type_names[0].replace('"', "").lower()
                        # NOTE strangely lower is required (Brazilian CTe)
                        self.implicit_many2ones[comodel_type].append(
                            (
                                self.filters.registry_name(
                                    klass.name, type_names=parent_names
                                ),
                                target_field,
                            )
                        )
                if klass not in self.all_complex_types:
                    self.all_complex_types.append(klass)

            for klass, path in all_file_classes:
                self._collect_extra_data(klass)

        package_dirs = set()

        # Generate modules
        for path, cluster in self.group_by_module(classes).items():
            should_skip_count = 0
            for pattern in SIGNATURE_CLASS_SKIP:
                for klass in cluster:
                    if re.search(pattern, klass.name):
                        should_skip_count += 1
                        break
            # a module/cluster may have some blacklisted classes
            # we want to skip the cluster only above some threshold
            if should_skip_count > 5:
                continue

            package_dirs.add(str(Path(path).parent))
            yield GeneratorResult(
                path=path.with_suffix(".py"),
                title=cluster[0].target_module,
                source=self.render_module(resolver, cluster),
            )

        self.config.output.max_line_length = 88  # OCA style
        self.ruff_code_oca(list(package_dirs))

    def render_module(
        self, resolver: DependenciesResolver, classes: List[Class]
    ) -> str:
        res = super().render_module(resolver, classes)

        # for some reason, when generating several files at once,
        # some field can loose their indention, we fix them here:
        field_prefix = self.filters.field_safe_prefix
        return "\n".join(
            [
                re.sub("^%s" % (field_prefix), "    %s" % (field_prefix), line)
                for line in res.splitlines()
            ]
        )

    def render_classes(
        self, classes: List[Class], module_namespace: Optional[str]
    ) -> str:
        """
        Render the source code of the classes.

        Overriden to control different number of line breaks for enums
        and classes for Odoo.
        """
        load = self.env.get_template

        def render_class(obj: Class) -> str:
            """Render class or enumeration."""
            if [ext.type for ext in obj.extensions]:
                # this is used only to change tag names, no Odoo model is required
                return ""
            if obj.is_enumeration:
                template = load("enum.jinja2")
            elif obj.is_service:
                template = load("service.jinja2")
            else:
                template = load("class.jinja2")

            string = template.render(
                obj=obj,
                module_namespace=module_namespace,
            ).strip()
            if not obj.is_enumeration:
                string = "\n" + string
            return string

        return "\n\n".join(map(render_class, classes)) + "\n"

    def _collect_extra_data(self, obj: Class):
        """Collect extra field data from the xsd file or another file"""

        location = (obj.location or "").replace("file://", "")
        if not os.path.isfile(location):
            return

        if not self.filters.files_to_etree.get(location):  # yes it can still happen
            xsd_tree = etree.parse(location)
            self.filters.files_to_etree[location] = xsd_tree
        else:
            xsd_tree = self.filters.files_to_etree[location]

        # if ComplexType has no description,
        # take it from the element declaration:
        if not obj.help:
            xpath_matches = xsd_tree.getroot().xpath(
                f"//xs:element[@name='{obj.name}']",
                namespaces={
                    "xs": "http://www.w3.org/2001/XMLSchema",
                    "xsd": "http://www.w3.org/2001/XMLSchema",
                },
            )
            if xpath_matches:
                children = xpath_matches[0].getchildren()
                if (
                    len(children) > 0
                    and children[0].tag
                    == "{http://www.w3.org/2001/XMLSchema}annotation"
                ):
                    obj.help = "".join(children[0].itertext()).strip()

        # extract fields choice attributes and types using xpath:
        for attr in obj.attrs:
            field_data = {}
            type_lookups = (
                f"//xs:element[@name='{obj.name}']//xs:element[@name='{attr.name}']",
                f"//xs:element[@name='{obj.name}']//xs:attribute[@name='{attr.name}']",
                f"//xs:complexType[@name='{obj.name}']//xs:element[@name='{attr.name}']",
                f"//xs:complexType[@name='{obj.name}']//xs:attribute[@name='{attr.name}']",
            )
            for lookup in type_lookups:
                xpath_matches = xsd_tree.getroot().xpath(
                    lookup,
                    namespaces={
                        "xs": "http://www.w3.org/2001/XMLSchema",
                        "xsd": "http://www.w3.org/2001/XMLSchema",
                    },
                )
                if xpath_matches:
                    xsd_choice_required = None
                    parent = xpath_matches[0].getparent()
                    # (here we don't try to group items by choice, but eventually we could)
                    while parent.tag == "{http://www.w3.org/2001/XMLSchema}sequence":
                        if (
                            parent.get("minOccurs", "1") == "0"
                        ):  # example veicTransp in Brazilian NFe
                            xsd_choice_required = False
                        parent = parent.getparent()
                    if parent.tag == "{http://www.w3.org/2001/XMLSchema}choice":
                        # here we assume only 1 choice per complexType
                        # but evexntually we could count them and number them...
                        field_data["choice"] = (
                            obj.name.lower()
                        )  # TODO consider changing choice -> xsd_choice
                        if (
                            parent.get("minOccurs", "1") != "0"
                            and xsd_choice_required is None
                        ):
                            # important feature we had in generateDS:
                            # if the element is part of <choice> tag without minOccurs="0"
                            # then it is required!
                            field_data["xsd_choice_required"] = True
                    xsd_type = xpath_matches[0].get("type")
                    if xsd_type and xsd_type not in [
                        "xsd:string",
                        "xsd:date",
                    ]:
                        field_data["xsd_type"] = xsd_type
                    self.filters.xsd_extra_info[f"{obj.name}#{attr.name}"] = field_data

    def ruff_code_oca(self, file_paths: list[str]) -> None:
        """
        Run a custom ruff format and check process on a list of files.

        This version is modified to:
        1. Format the code first.
        2. Check and fix for both 'F' (Pyflakes, includes unused imports)
        and 'I' (isort) rules.
        """
        if not file_paths:
            return

        # First, run ruff format to handle all styling.
        format_command = [
            "ruff",
            "format",
            "--config",
            f"line-length={self.config.output.max_line_length}",
            *file_paths,
        ]

        # Next, run ruff check to fix code issues, including unused imports.
        # We select both 'F' and 'I' rules.
        check_command = [
            "ruff",
            "check",
            "--config",
            f"line-length={self.config.output.max_line_length}",
            "--select",  # Use --select instead of --config for cleaner syntax
            "F,I",  # Selects both F (Pyflakes) and I (isort) rules
            "--fix",
            "--unsafe-fixes",  # Often needed to remove unused imports
            "--exit-zero",  # Don't fail the build if there are unfixable errors
            *file_paths,
        ]

        try:
            # Run formatting first
            subprocess.run(format_command, capture_output=True, check=True, text=True)
            # Then run linting and fixing
            subprocess.run(check_command, capture_output=True, check=True, text=True)
        except subprocess.CalledProcessError as e:
            details = e.stderr.strip().replace("error: ", "")
            raise CodegenError(f"Ruff failed\n{details}")
        except FileNotFoundError:
            raise CodegenError(
                "Ruff command not found. Is it installed and in your PATH?"
            )
